import os
import warnings

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import StratifiedKFold
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import LinearSVC
from sklearn.tree import DecisionTreeClassifier

from classification.feature_select import FeatureSelection
from utils.data_process import DataProcessing

warnings.filterwarnings('ignore')


class Classification:
    def __init__(self, pro_name='COAD'):
        self.pro_name = pro_name
        self.dp = DataProcessing(pro_name=pro_name)
        self.fs = FeatureSelection(pro_name=pro_name)
        self.matrix = self.dp.get_classification_matrix()
        self.label = self.dp.get_raw_label()
        self.clf_data_path = '../dataset/clf_data/'
        self.clf_record_path = '../record/clf_record/'
        self.classifier = {'LinearSVC': LinearSVC(C=1.0, penalty='l1', dual=False, random_state=1),
                           'KNN': KNeighborsClassifier(n_neighbors=5), 'GaussianNB': GaussianNB(),
                           'DecisionTreeClassifier': DecisionTreeClassifier(random_state=1),
                           'LogisticRegression': LogisticRegression(),
                           'RandomForestClassifier': RandomForestClassifier(n_estimators=10)}

    def classify(self, fs_name='TriVote', clf_name='LogisticRegression', k=70):
        sf = getattr(self.fs, self.fs.select_function[fs_name])
        matrix = sf(k)
        label = self.fs.label
        kf = StratifiedKFold(n_splits=10, random_state=1, shuffle=True)  # 每次分组都相同
        acc_list = []
        for train, val in kf.split(matrix, label):
            train_data, train_label = np.array(matrix)[train], np.array(label)[train]
            val_data, val_label = np.array(matrix)[val], np.array(label)[val]
            clf = self.classifier[clf_name]
            clf.fit(train_data, train_label)
            predicts = clf.predict(val_data)
            error = 0
            for i in range(len(predicts)):
                if predicts[i] != val_label[i]:
                    error += 1
            acc = round((len(predicts) - error) / len(predicts), 6)
            acc_list.append(acc)
        return matrix, matrix.shape[1], np.array(acc_list).mean()

    def find_best_feanum(self):
        feanum_list = np.arange(30, 500, 20)
        for fs_name in self.fs.select_function.keys():
            best_matrix = self.fs.matrix
            best_feanum = 0
            best_acc = 0
            for k in feanum_list:
                matrix, feanum, acc = self.classify(fs_name=fs_name, clf_name='LogisticRegression', k=k)
                if acc > best_acc:
                    best_matrix = matrix
                    best_feanum = feanum
                    best_acc = acc
            with open(self.clf_data_path + self.pro_name + '_' + fs_name + '_' + str(best_feanum) + '.txt',
                      'w') as file:
                for i in range(best_matrix.shape[0]):
                    for j in range(best_matrix.shape[1]):
                        file.write(str(best_matrix[i][j]) + '\t')
                    file.write('\n')
            print(self.pro_name + '_' + fs_name + '_' + str(best_feanum) + '.txt is saved! ')

    def get_record(self):
        file_list = os.listdir(self.clf_data_path)
        file_path = self.clf_data_path
        record = []
        for fs_name in self.fs.select_function.keys():
            record.append([])
            for file_name in file_list:
                if (self.pro_name in file_name) and (fs_name in file_name):
                    file_path = self.clf_data_path + file_name
            matrix = pd.read_table(file_path, sep='\t', header=None, low_memory=False).values
            label = self.fs.label
            matrix = matrix[:, :-1]
            for clf_name in self.classifier.keys():
                kf = StratifiedKFold(n_splits=10, random_state=1, shuffle=True)
                acc_list = []
                for train, val in kf.split(matrix, label):
                    train_data, train_label = np.array(matrix)[train], np.array(label)[train]
                    val_data, val_label = np.array(matrix)[val], np.array(label)[val]
                    clf = self.classifier[clf_name]
                    clf.fit(train_data, train_label)
                    predicts = clf.predict(val_data)
                    error = 0
                    for i in range(len(predicts)):
                        if predicts[i] != val_label[i]:
                            error += 1
                    acc = round((len(predicts) - error) / len(predicts), 6)
                    acc_list.append(acc)
                record[-1].append(np.array(acc_list).mean())
        record = np.array(record)
        with open(self.clf_record_path + self.pro_name + '.txt', 'w') as file:
            for i in range(record.shape[0]):
                for j in range(record.shape[1]):
                    file.write(str(record[i][j]) + '\t')
                file.write('\n')
        print(self.pro_name + ' record is saved! ')

