import gc
import time
import warnings

import numpy as np
import torch
import torch.optim as optim

from regression.models import ae, ae_wgan, my_model, s_ae_wgan, sae, sae_sgan, sae_wgan, vae, vae_wgan
from regression.models.model import AE, VAE, Discriminator, MA_Discriminator
from utils.data_process import DataProcessing
from sklearn.model_selection import KFold

warnings.filterwarnings('ignore')
np.set_printoptions(formatter={'float': '{: .2f}'.format})


class FeatureConstruction:
    def __init__(self, max_epochs=3000, min_epochs=20, batch_size=20, lr=1e-4, n1=400, n2=400, gradient=200,
                 num_heads=10, pro_name='COAD', model_name='AE_SGAN'):
        self.pro_name = pro_name
        self.model_name = model_name
        self.dp = DataProcessing(pro_name, model_name)
        self.fs_data_path = 'dataset/fs_data/'
        self.fc_data_path = 'dataset/fc_data/'
        self.model_save_path = 'regression/model_weight/'
        self.max_epochs, self.min_epochs = max_epochs, min_epochs
        self.batch_size = batch_size
        self.lr = lr
        self.n1, self.n2, self.gradient, self.num_heads = n1, n2, gradient, num_heads
        self.model_list = {'VAE': VAE,
                           'SAE': AE,
                           'AE': AE,

                           'AE_WGAN': AE,
                           'SAE_WGAN': AE,
                           'VAE_WGAN': VAE,

                           'MA_MyModel': AE,
                           'MyModel': AE,
                           'MA_SAE_SGAN': AE,
                           'MA_S_AE_WGAN': AE,
                           'MA_SAE_WGAN': AE
                           }

    # 早停策略
    def earlyStop(self, loss_list, epoch):
        if epoch < self.min_epochs:
            return False
        elif (epoch == self.max_epochs - 1):
            return True
        elif (loss_list[epoch] > loss_list[epoch - 1]):
            return True
        elif (epoch > 4) and ('%.4f' % loss_list[epoch] == '%.4f' % loss_list[epoch - 1]) and (
                '%.4f' % loss_list[epoch - 1] == '%.4f' % loss_list[epoch - 2]) and (
                '%.4f' % loss_list[epoch - 2] == '%.4f' % loss_list[epoch - 3]) and (
                '%.4f' % loss_list[epoch - 3] == '%.4f' % loss_list[epoch - 4]):
            return True
        else:
            return False

    def construct(self, i):
        if ('GAN' in self.model_name) or ('MyModel' in self.model_name):
            loss_list = self.gan_construct(i)
        else:
            loss_list = self.original_construct(i)
        return loss_list

    def original_construct(self, i):
        loss_summary = []
        np.set_printoptions(precision=3)
        if self.model_name == 'VAE':
            base_model = vae.VAE(self.max_epochs, self.batch_size, self.lr)
        elif self.model_name == 'SAE':
            base_model = sae.SAE(self.max_epochs, self.batch_size, self.lr)
        else:
            base_model = ae.AE(self.max_epochs, self.batch_size, self.lr)

        matrix = self.dp.get_fc_inp(i)
        days, m = self.dp.get_days_standardScaler()
        kf = KFold(n_splits=10, shuffle=True, random_state=6)
        test_matrix_temp = []
        test_index_temp = []
        count = 0
        for train_index,test_index in kf.split(matrix,days):
            train_matrix,train_days=matrix[train_index],days[train_index]
            test_matrix,test_days=matrix[test_index],days[test_index]
            train_dataLoader, train_data_size = self.dp.getLoader(train_matrix, train_days, self.batch_size)
            test_dataLoader,test_data_size = self.dp.getLoader(test_matrix,test_days,self.batch_size)
            loss_list = np.zeros((self.max_epochs))
            net = self.model_list[self.model_name](self.n1, self.n2)
            optimizer = optim.Adam(net.parameters(), lr=self.lr)
            net, optimizer = base_model.pre_train(train_dataLoader, net, optimizer)
            for epoch in range(self.max_epochs):
                train_loss, net, optimizer, hid1, hid2, out, order_list = base_model.train(train_dataLoader,
                                                                                           train_data_size,
                                                                                           net, optimizer)
                loss_list[epoch] = format(train_loss, '.4f')
                if self.earlyStop(loss_list, epoch):
                    _, hid1, hid2, out, order_list = base_model.test(test_dataLoader, test_data_size, net)
                    test_matrix = self.dp.combineOut(test_matrix, hid1, hid2, out, order_list)
                    test_matrix_temp.extend(test_matrix)
                    test_index_temp.extend(test_index)
                    print('[loss]: {}'.format(loss_list[epoch]))
                    loss_summary.append([str(loss_list[epoch]), str(i)])
                    torch.save(net.state_dict(),
                               self.model_save_path + self.model_name + '_' + self.pro_name + '_' + str(
                                   i)+ '_' +str(count) + '.pkl')
                    count+=1
                    del train_dataLoader, net, optimizer
                    gc.collect()
                    time.sleep(5)
                    break
                else:
                    # print('[epoch]: {}\t[loss]: {}'.format(epoch, format(train_loss, '.4f')))
                    pass
        test_matrix_temp = np.array(test_matrix_temp)
        test_index_temp = np.array(test_index_temp)
        _, test_matrix_temp = zip(*sorted(zip(test_index_temp, test_matrix_temp)))
        test_matrix_temp = np.array(test_matrix_temp)
        with open(self.fs_data_path + self.model_name + '_' + self.pro_name + '_' + str(i + 1) + '.txt',
                  'w+') as tf:
            for a in range(test_matrix_temp.shape[0]):
                for b in range(test_matrix_temp.shape[1]):
                    tf.write(str(test_matrix_temp[a][b]) + '\t')
                tf.write('\n')
        return np.array(loss_summary)

    def gan_construct(self, i):
        np.set_printoptions(precision=3)
        loss_summary = []
        name_list = self.model_name.split('_')
        if 'WGAN' in name_list:
            if ('AE' in name_list) and ('S' in name_list):
                base_model = s_ae_wgan.S_AEWGAN(self.max_epochs, self.batch_size, self.lr)
            elif 'VAE' in name_list:
                base_model = vae_wgan.VAEWGAN(self.max_epochs, self.batch_size, self.lr)
            elif 'SAE' in name_list:
                base_model = sae_wgan.SAEWGAN(self.max_epochs, self.batch_size, self.lr)
            else:  # AE
                base_model = ae_wgan.AEWGAN(self.max_epochs, self.batch_size, self.lr)
        elif 'SGAN' in name_list:
            base_model = sae_sgan.SAESGAN(self.max_epochs, self.batch_size, self.lr)
        else:
            base_model = my_model.MyModel(self.max_epochs, self.batch_size, self.lr)
        matrix = self.dp.get_fc_inp(i)
        days, m = self.dp.get_days_standardScaler()
        kf = KFold(n_splits=10,shuffle=True,random_state=6)
        test_matrix_temp=[]
        test_index_temp=[]
        count=0
        for train_index,test_index in kf.split(matrix,days):
            train_matrix,train_days=matrix[train_index],days[train_index]
            test_matrix,test_days=matrix[test_index],days[test_index]
            train_dataLoader, train_data_size = self.dp.getLoader(train_matrix, train_days, self.batch_size)
            test_dataLoader,test_data_size = self.dp.getLoader(test_matrix,test_days,self.batch_size)
            loss_list = np.zeros((self.max_epochs))
            generator = self.model_list[self.model_name](self.n1, self.n2)
            if 'MA' in self.model_name:
                discriminator = MA_Discriminator(generator, self.n1, self.n2, self.gradient, self.num_heads)
            else:
                discriminator = Discriminator(generator, self.n1, self.n2)
            g_optimizer = optim.RMSprop(generator.parameters(), lr=self.lr)
            d_optimizer = optim.RMSprop(discriminator.parameters(), lr=self.lr)
            generator, g_optimizer = base_model.pre_train(train_dataLoader, generator, g_optimizer)
            for epoch in range(self.max_epochs):
                train_loss, generator, g_optimizer, discriminator, d_optimizer, hid1, hid2, out, order_list = \
                    base_model.train(train_dataLoader, train_data_size, generator, g_optimizer,
                                     discriminator, d_optimizer)
                loss_list[epoch] = format(train_loss, '.4f')
                if self.earlyStop(loss_list, epoch):
                    _, hid1, hid2, out, order_list=base_model.test(test_dataLoader,test_data_size,generator,discriminator)
                    test_matrix=self.dp.combineOut(test_matrix,hid1,hid2,out,order_list)
                    test_matrix_temp.extend(test_matrix)
                    test_index_temp.extend(test_index)
                    print('[loss]: {}'.format(loss_list[epoch]))
                    loss_summary.append([loss_list[epoch], i])
                    torch.save(generator.state_dict(),
                               self.model_save_path + self.model_name + '_' + self.pro_name + '_' + str(
                                   i)+ '_' +str(count) + '.pkl')
                    count+=1
                    del train_dataLoader, generator, g_optimizer, discriminator, d_optimizer
                    gc.collect()
                    time.sleep(5)
                    break
                else:
                    # print('[epoch]: {}\t[train_loss]: {}'.format(epoch, format(train_loss, '.4f')))
                    pass
        test_matrix_temp=np.array(test_matrix_temp)
        test_index_temp=np.array(test_index_temp)
        _,test_matrix_temp=zip(*sorted(zip(test_index_temp,test_matrix_temp)))
        test_matrix_temp=np.array(test_matrix_temp)
        with open(self.fs_data_path + self.model_name + '_' + self.pro_name + '_' + str(i + 1) + '.txt',
                  'w+') as tf:
            for a in range(test_matrix_temp.shape[0]):
                for b in range(test_matrix_temp.shape[1]):
                    tf.write(str(test_matrix_temp[a][b]) + '\t')
                tf.write('\n')
        return np.array(loss_summary)
