import warnings

import numpy as np
import pandas as pd
import torch
from sklearn.metrics import mean_absolute_error
from sklearn.model_selection import KFold
from sklearn.preprocessing import StandardScaler
from sklearn.svm import LinearSVR
from torch.utils.data import DataLoader, TensorDataset

warnings.filterwarnings('ignore')


class DataProcessing:
    def __init__(self, pro_name='COAD', model_name='VAE'):
        self.pro_name = pro_name
        self.model_name = model_name
        self.raw_matrix5000_path = '../dataset/raw_data/matrix_5000/'
        self.final_matrix_path = 'dataset/raw_data/final_matrix/'
        self.final_label_path = '../dataset/raw_data/final_label/'
        self.final_days_path = 'dataset/raw_data/final_days/'
        self.clf_data_path = '../dataset/clf_data/'
        self.fs_data_path = 'dataset/fs_data/'
        self.fc_data_path = 'dataset/fc_data/'

    def get_regression_matrix(self):
        matrix = pd.read_table(self.final_matrix_path + self.pro_name + '.txt', sep='\t', header=None,
                               low_memory=False).values
        matrix = matrix[:, :-1]
        return matrix

    def get_classification_matrix(self):
        matrix = pd.read_table(self.raw_matrix5000_path + self.pro_name + '.txt', sep='\t', header=None,
                               low_memory=False).values
        matrix = matrix[np.argsort(matrix[:, 0])]
        matrix = matrix[1:, 1:-1]
        return matrix

    def get_raw_days(self):
        days = pd.read_table(self.final_days_path + self.pro_name + '.txt', header=None).values
        return days

    def get_raw_label(self):
        label = pd.read_table(self.final_label_path + self.pro_name + '.txt', header=None).values
        label = label[1:, 1]
        return label

    def get_days_standardScaler(self):
        days = self.get_raw_days()
        m = StandardScaler()
        days = m.fit_transform(days)
        days = days.squeeze()
        return days, m

    def evaluate(self, matrix_train, days_train, matrix, days, m):
        lsvr = LinearSVR()
        lsvr.fit(matrix_train, days_train)
        predict = lsvr.predict(matrix)
        predict, days = m.inverse_transform(predict), m.inverse_transform(days)
        return mean_absolute_error(predict, days)

    def cross_validation(self, matrix, days, m, n_splits=10):
        lsvr_mae = []
        kf = KFold(n_splits=n_splits, random_state=self.random_seed1, shuffle=True)
        for train, val in kf.split(matrix, days):
            t_data, t_label = np.array(matrix)[train], np.array(days)[train]
            v_data, v_label = np.array(matrix)[val], np.array(days)[val]
            mae = self.evaluate(t_data, t_label, v_data, v_label, m)
            lsvr_mae.append(mae)
        lsvr_mae = np.array(lsvr_mae)
        return lsvr_mae.mean()

    def get_first_fs(self):
        matrix = pd.read_table(self.fc_data_path + self.pro_name + '.txt', sep='\t', header=None).values
        days, m = self.get_days_standardScaler()
        matrix = matrix[:, :-1]
        mae = self.cross_validation(matrix, days, m)
        return mae, matrix.shape[1]

    def getLoader(self, matrix, label, batch_size):
        size = matrix.shape[0]
        label = label.reshape(-1, 1)
        order = np.arange(0, size)
        order = torch.tensor(order)
        order = order.to(torch.float32)
        matrix = torch.tensor(matrix)
        matrix = matrix.to(torch.float32)
        label = torch.tensor(label)
        label = label.to(torch.float32)
        dataset = TensorDataset(order, label, matrix)
        dataLoader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=True)
        return dataLoader, size

    def outToArray(self, out_list):
        out_list = [np.array(item.detach()) for item in out_list]
        out_list = np.array(out_list)
        temp = []
        for item in out_list:
            for t in item:
                temp.append(t)
        temp = np.array(temp)
        return temp

    def combineOut(self, input, hid1, hid2, out, order_list):
        hid1, hid2, out, order_list = self.outToArray(hid1), self.outToArray(hid2), self.outToArray(
            out), self.outToArray(order_list)
        z = zip(order_list, hid1, hid2, out)
        z = sorted(z)
        order_list, hid1, hid2, out = zip(*z)
        return np.hstack((input, hid1, hid2, out))

    def get_fc_inp(self, i):
        if i == 0:
            matrix = pd.read_table(self.fc_data_path + self.pro_name + '.txt', sep='\t', header=None).values
        else:
            matrix = pd.read_table(self.fc_data_path + self.model_name + '_' + self.pro_name + '_' + str(i) + '.txt',
                                   sep='\t', header=None).values
        matrix = matrix[:, :-1]
        return matrix

    def get_fs_inp(self, i):
        matrix = pd.read_table(self.fs_data_path + self.model_name + '_' + self.pro_name + '_' + str(i) + '.txt',
                               sep='\t', header=None).values
        days, m = self.get_days_standardScaler()
        matrix = matrix[:, :-1]
        return matrix, days, len(days), m
